-----------------------------
-- Faster hangable ceiling --
-----------------------------

-- Original Code by Sunk

---@param num number
---@return number
local function convert_s16(num)
	num = num & 0xFFFF
	return ((num >= 0x7FFF) and (num - 0x10000) or num)
end

---@param m MarioState
local function update_custom_hang_moving(m)
	local stepResult = 0
	local nextPos = {}
	local maxSpeed = 25

	m.forwardVel = m.forwardVel + 16
	if m.forwardVel > maxSpeed then
		m.forwardVel = maxSpeed
	end

	m.faceAngle.y = m.intendedYaw - approach_s32(convert_s16(m.intendedYaw - m.faceAngle.y), 0, 0x800, 0x800)

	m.slideYaw = m.faceAngle.y
	m.slideVelX = m.forwardVel * sins(m.faceAngle.y)
	m.slideVelZ = m.forwardVel * coss(m.faceAngle.y)

	m.vel.x = m.slideVelX
	m.vel.y = 0.0
	m.vel.z = m.slideVelZ

	nextPos.x = m.pos.x - m.ceil.normal.y * m.vel.x
	nextPos.z = m.pos.z - m.ceil.normal.y * m.vel.z
	nextPos.y = m.pos.y

	stepResult = perform_hanging_step(m, nextPos)

	vec3f_copy(m.marioObj.header.gfx.pos, m.pos)
	vec3s_set(m.marioObj.header.gfx.angle, 0, m.faceAngle.y, 0)
	return stepResult
end

---@param m MarioState
function act_custom_hang_moving(m)
	if m.input & INPUT_A_DOWN == 0 then
		return set_mario_action(m, ACT_FREEFALL, 0)
	end

	if m.input & INPUT_Z_PRESSED ~= 0 then
		return set_mario_action(m, ACT_GROUND_POUND, 0)
	end

	if m.ceil == nil or m.ceil.type ~= SURFACE_HANGABLE then
		return set_mario_action(m, ACT_FREEFALL, 0)
	end

	if m.actionArg & 1 ~= 0 then
		set_mario_animation(m, MARIO_ANIM_MOVE_ON_WIRE_NET_RIGHT)
	else
		set_mario_animation(m, MARIO_ANIM_MOVE_ON_WIRE_NET_LEFT)
	end

	if m.marioObj.header.gfx.animInfo.animFrame == 12 then
		play_sound(SOUND_ACTION_HANGING_STEP, m.marioObj.header.gfx.cameraToObject)
		queue_rumble_data_mario(m, 5, 30)
	end

	if is_anim_past_end(m) ~= 0 then
		m.actionArg = m.actionArg ~ 1
		if m.input & INPUT_ZERO_MOVEMENT ~= 0 then
			return set_mario_action(m, ACT_HANGING, m.actionArg)
		end
	end

	if update_custom_hang_moving(m) == 2 --[[HANG_LEFT_CEIL]] then
		set_mario_action(m, ACT_FREEFALL, 0)
	end

	return 0
end

--------------------------------------------------------------------------------------------------------

local actions_able_to_wallkick =
{
	[ACT_JUMP] = ACT_JUMP,
	[ACT_HOLD_JUMP] = ACT_HOLD_JUMP,
	[ACT_DOUBLE_JUMP] = ACT_DOUBLE_JUMP,
	[ACT_TRIPLE_JUMP] = ACT_TRIPLE_JUMP,
	[ACT_SIDE_FLIP] = ACT_SIDE_FLIP,
	[ACT_BACKFLIP] = ACT_BACKFLIP,
	[ACT_LONG_JUMP] = ACT_LONG_JUMP,
	[ACT_WALL_KICK_AIR] = ACT_WALL_KICK_AIR,
	[ACT_TOP_OF_POLE_JUMP] = ACT_TOP_OF_POLE_JUMP,
	[ACT_FREEFALL] = ACT_FREEFALL
}

--This is in degrees
gGlobalSyncTable.limit = 46

--Thanks Djoslin
function convert_s16(num)
	local min = -32768
	local max = 32767
	while (num < min) do
		num = max + (num - min)
	end
	while (num > max) do
		num = min + (num - max)
	end
	return num
end

--This is mostly copied from the wall bonk check code
---@param m MarioState
function wallkicks(m)
	if m.playerIndex ~= 0 then return end

	if m.wall ~= nil then
		if (m.wall.type == SURFACE_BURNING) then return end

		local wallDYaw = (atan2s(m.wall.normal.z, m.wall.normal.x) - (m.faceAngle.y))
		--I don't really understand this however I do know the lower `limit` becomes, the more possible wallkick degrees.
		local limitNegative = (-((180 - gGlobalSyncTable.limit) * (8192/45))) + 1
		local limitPositive = ((180 - gGlobalSyncTable.limit) * (8192/45)) - 1
		--wallDYaw is s16, so I converted it
		wallDYaw = convert_s16(wallDYaw)

		--Standard air hit wall requirements
		if (m.forwardVel >= 16) and (actions_able_to_wallkick[m.action] ~= nil) then
			if (wallDYaw >= limitPositive) or (wallDYaw <= limitNegative) then
				mario_bonk_reflection(m, 0)
				m.faceAngle.y = m.faceAngle.y + 0x8000
				set_mario_action(m, ACT_AIR_HIT_WALL, 0)
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------

function mario_on_set_action(m)
	if m.action == ACT_JUMP or m.action == ACT_DOUBLE_JUMP or
		m.action == ACT_TRIPLE_JUMP or m.action == ACT_SPECIAL_TRIPLE_JUMP or
		m.action == ACT_STEEP_JUMP or m.action == ACT_SIDE_FLIP or
		m.action == ACT_BACKFLIP or m.action == ACT_LONG_JUMP or
		m.action == ACT_RIDING_SHELL_JUMP then 
		if m.pos.y < m.floorHeight - 6 then
			m.pos.y = m.pos.y +5
			m.vel.y = m.vel.y +11
		end
	end
end

--------------------------------------------------------------------------------------------------------

---@param obj Object
function bhv_custom_pushable_loop(obj)
	local m = nearest_mario_state_to_object(obj)
	local player = m.marioObj

	if obj_check_if_collided_with_object(obj, player) == 1 and m ~= nil and m.flags & MARIO_UNKNOWN_31 ~= 0 then
		if check_if_moving_over_floor(8, 150) == 1 then
			obj.oForwardVel = 10
			cur_obj_move_using_fvel_and_gravity()
		end
	end
end

hook_behavior(id_bhvPushableMetalBox, OBJ_LIST_SURFACE, false, nil, bhv_custom_pushable_loop, "bhvPushableMetalBox")

---@param m MarioState
local function mario_update(m)
	if m.playerIndex ~= 0 then return end

	if m.flags & MARIO_UNKNOWN_31 ~= 0 then
		mario_set_forward_vel(m, 15)
	end
end

--------------------------------------------------------------------------------------------------------

function act_custom_ledge_grab(m)
	local intendedDYaw = m.intendedYaw - m.faceAngle.y
	local hasSpaceForMario = (m.ceilHeight - m.floorHeight >= 160)

	if (m.actionTimer < 10) then
		m.actionTimer = m.actionTimer + 1
	end

	if (m.input & (INPUT_Z_PRESSED | INPUT_OFF_FLOOR) ~= 0) then
		return let_go_of_ledge(m)
	end

	if ((m.input & INPUT_A_PRESSED) ~= 0 and hasSpaceForMario) then
		return set_mario_action(m, ACT_LEDGE_CLIMB_FAST, 0)
	end

	if (m.input & INPUT_UNKNOWN_10) ~= 0 then
		return let_go_of_ledge(m)
	end

	if (m.actionTimer == 10 and (m.input & INPUT_NONZERO_ANALOG) ~= 0) then
		if (intendedDYaw >= -0x4000 and intendedDYaw <= 0x4000) then
			if (hasSpaceForMario) then
				return set_mario_action(m, ACT_LEDGE_CLIMB_SLOW_1, 0)
			end
		else
			return let_go_of_ledge(m)
		end
	end

	local heightAboveFloor = m.pos.y - find_floor_height_relative_polar(m, -0x8000, 30)
	if (hasSpaceForMario and heightAboveFloor < 100) then
		return set_mario_action(m, ACT_LEDGE_CLIMB_FAST, 0)
	end

	if (m.actionArg == 0) then
		play_character_sound_if_no_flag(m, CHAR_SOUND_WHOA, MARIO_MARIO_SOUND_PLAYED)
	end

	stop_and_set_height_to_floor(m)
	set_character_animation(m, CHAR_ANIM_IDLE_ON_LEDGE)

	return 0
end

hook_event(HOOK_BEFORE_PHYS_STEP, wallkicks)
hook_event(HOOK_ON_SET_MARIO_ACTION, mario_on_set_action)
hook_event(HOOK_MARIO_UPDATE, mario_update)

hook_mario_action(ACT_HANG_MOVING, act_custom_hang_moving)
hook_mario_action(ACT_LEDGE_GRAB, act_custom_ledge_grab)